#
# MIT License
#
# Copyright (c) 2023 Mike Heddes, Igor Nunes, Pere Vergés, Denis Kleyko, and Danny Abraham
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import math
from typing import Optional
import torch
from torch import Tensor
import torch.nn as nn
from scipy.stats import binom

import torchhd.functional as functional
from torchhd.tensors.base import VSATensor

__all__ = [
    "SparseDistributed",
    "hopfield",
    "Hopfield",
    "modern_hopfield",
    "attention",
]


class SparseDistributed(nn.Module):
    r"""`Sparse Distributed Memory <https://redwood.berkeley.edu/wp-content/uploads/2020/08/KanervaP_SDMrelated_models1993.pdf>`_

    The Sparse Distributed Memory (SDM) is specified by its (typically random) keys and their values.

    Args:
        memory_size (int): The number of memory key-value pairs.
        key_dim (int): The dimensionality of the key vectors.
        value_dim (int): The dimensionality of the value vectors.
        p (float, optional): The expected fraction of memory address that will contain any value. Default: ``0.000368``.
        kappa (int, optional): The maximum count for each memory cell, values are clipped between [-kappa, kappa]. Default: no clipping.
        dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None`` depends on VSATensor.
        device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Shapes:
        - Keys: :math:`(n, a)`
        - Values: :math:`(n, c)`

    Examples::
        >>> keys = torchhd.random(6, 512)
        >>> sdm = torchhd.memory.SparseDistributed(100000, 512, 512)
        >>> # use as associative memory
        >>> sdm.write(keys, keys)
        >>> read = sdm.read(keys).sign()
        >>> torchhd.cosine_similarity(read, keys)
        tensor([[ 1.0000,  0.0156, -0.0039, -0.0742,  0.0000, -0.0195],
                [ 0.0156,  1.0000, -0.0352, -0.0586,  0.0000, -0.0039],
                [-0.0039, -0.0352,  1.0000,  0.0156,  0.0820, -0.0234],
                [-0.0742, -0.0586,  0.0156,  1.0000, -0.0039,  0.0000],
                [ 0.0000,  0.0000,  0.0820, -0.0039,  1.0000,  0.0195],
                [-0.0195, -0.0039, -0.0234,  0.0000,  0.0195,  1.0000]])
    """

    memory_size: int
    key_dim: int
    value_dim: int
    keys: VSATensor
    values: VSATensor
    threshold: int
    kappa: Optional[int]

    def __init__(
        self,
        memory_size: int,
        key_dim: int,
        value_dim: int,
        p: float = 0.000368,
        kappa: Optional[int] = None,
        dtype=None,
        device=None,
        requires_grad=False,
    ) -> None:
        super().__init__()

        self.memory_size = memory_size
        self.key_dim = key_dim
        self.value_dim = value_dim

        radius = int(binom.ppf(p, key_dim, 0.5))
        self.threshold = key_dim - 2 * radius
        self.kappa = kappa

        keys = functional.random(memory_size, key_dim, dtype=dtype, device=device)
        self.keys = nn.Parameter(keys, requires_grad)

        values = functional.empty(memory_size, value_dim, device=device, dtype=dtype)
        self.values = nn.Parameter(values, requires_grad)

    def read(self, query: Tensor) -> VSATensor:
        r"""Read value from Sparse Distributed Memory whose key is most similar to the query.

        Args:
            query (Tensor): The query vector for the memory lookup.

        Shapes:
            - Query: :math:`(*, d)`
            - Result: :math:`(*, d)`

        """
        # first dims from query, last dim from value
        out_shape = (*query.shape[:-1], self.value_dim)

        if query.dim() == 1:
            query = query.unsqueeze(0)

        # make sure to have at least two dimension for index_add_
        intermediate_shape = (*query.shape[:-1], self.value_dim)

        similarity = query @ self.keys.T
        is_active = similarity >= self.threshold

        # sparse matrix-vector multiplication
        r_indices, v_indices = is_active.nonzero().T
        read = query.new_zeros(intermediate_shape)
        read.index_add_(0, r_indices, self.values[v_indices])
        return read.view(out_shape)

    @torch.no_grad()
    def write(self, keys: Tensor, values: Tensor) -> None:
        r"""Write value to Sparse Distributed Memory at address.

        Args:
            address (Tensor): The address vector for the write to memory.
            value (Tensor): The value vector written to memory.

        Shapes:
            - Address: :math:`(*, d)`
            - Value: :math:`(*, d)`

        """

        if keys.dim() == 1:
            keys = keys.unsqueeze(0)

        if values.dim() == 1:
            values = values.unsqueeze(0)

        similarity = keys @ self.keys.T
        is_active = similarity >= self.threshold

        # sparse outer product and addition
        from_indices, to_indices = is_active.nonzero().T
        self.values.index_add_(0, to_indices, values[from_indices])

        if self.kappa is not None:
            self.values.clamp_(-self.kappa, self.kappa)


def hopfield(query: Tensor, memory: Tensor, kappa: int = None) -> Tensor:
    r"""`Classical Hopfield network <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC346238/>`_

    Args:
        query (Tensor): The query vector for the memory lookup.
        memory (Tensor): The items of memory for the memory lookup.

    Shapes:
        - Query: :math:`(*, d)`
        - Memory: :math:`(n, d)`
        - Result: :math:`(*, d)`

    Examples::
        >>> items = torchhd.random(6, 512)
        >>> read = memory.hopfield(items, items).sign()
        >>> torchhd.cosine_similarity(read, items)
        tensor([[ 1.0000,  0.0156, -0.0039, -0.0742,  0.0000, -0.0195],
                [ 0.0156,  1.0000, -0.0352, -0.0586,  0.0000, -0.0039],
                [-0.0039, -0.0352,  1.0000,  0.0156,  0.0820, -0.0234],
                [-0.0742, -0.0586,  0.0156,  1.0000, -0.0039,  0.0000],
                [ 0.0000,  0.0000,  0.0820, -0.0039,  1.0000,  0.0195],
                [-0.0195, -0.0039, -0.0234,  0.0000,  0.0195,  1.0000]])


    """
    product = memory.T @ memory
    torch.diagonal(product).zero_()

    if kappa is not None:
        product.clamp_(-kappa, kappa)

    return query @ product


class Hopfield(nn.Module):
    r"""`Classical Hopfield network <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC346238/>`_

    Args:
        vector_dim (int): The dimensionality of the vectors in the memory.
        kappa (int, optional): The maximum count for each memory cell, values are clipped between [-kappa, kappa]. Default: no clipping.
        dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None`` depends on VSATensor.
        device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Shapes:
        - Memory: :math:`(d, d)`

    Examples::
        >>> items = torchhd.random(6, 512)
        >>> hopfield = torchhd.memory.Hopfield(512)
        >>> hopfield.write(items)
        >>> read = hopfield.read(items).sign()
        >>> torchhd.cosine_similarity(read, items)
        tensor([[ 1.0000,  0.0156, -0.0039, -0.0742,  0.0000, -0.0195],
                [ 0.0156,  1.0000, -0.0352, -0.0586,  0.0000, -0.0039],
                [-0.0039, -0.0352,  1.0000,  0.0156,  0.0820, -0.0234],
                [-0.0742, -0.0586,  0.0156,  1.0000, -0.0039,  0.0000],
                [ 0.0000,  0.0000,  0.0820, -0.0039,  1.0000,  0.0195],
                [-0.0195, -0.0039, -0.0234,  0.0000,  0.0195,  1.0000]])
    """

    vector_dim: int
    memory: Tensor
    kappa: Optional[int]

    def __init__(
        self,
        vector_dim: int,
        kappa: Optional[int] = None,
        dtype=None,
        device=None,
        requires_grad=False,
    ) -> None:
        super().__init__()

        self.vector_dim = vector_dim
        self.kappa = kappa

        memory = torch.zeros(
            self.vector_dim, self.vector_dim, device=device, dtype=dtype
        )
        self.memory = nn.Parameter(memory, requires_grad)

    def read(self, query: Tensor) -> Tensor:
        r"""Read value from Hopfield network at key most similar to the query.

        Args:
            query (Tensor): The query vector for the memory lookup.

        Shapes:
            - Query: :math:`(*, d)`
            - Result: :math:`(*, d)`

        """
        return query @ self.memory.T

    @torch.no_grad()
    def write(self, items: Tensor) -> Tensor:
        r"""Write items to Hopfield Memory.

        Args:
            items (Tensor): The item vectors to write to memory.

        Shapes:
            - Items: :math:`(*, d)`

        """

        if items.dim() == 1:
            items = items.unsqueeze(0)

        # Add the outer product to memory
        self.memory.add_(items.T @ items)
        torch.diagonal(self.memory).zero_()

        if self.kappa is not None:
            self.memory.clamp_(-self.kappa, self.kappa)


def modern_hopfield(query: Tensor, memory: Tensor) -> Tensor:
    r"""`Modern Hopfield network <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC346238/>`_

    Also known as Dense Associative Memory.

    Args:
        query (Tensor): The query vector for the memory lookup.
        memory (Tensor): The items of memory for the memory lookup.

    Shapes:
        - Query: :math:`(*, d)`
        - Memory: :math:`(n, d)`
        - Result: :math:`(*, d)`


    Examples::
        >>> items = torchhd.random(6, 512)
        >>> read = memory.dense_associative(items, items).sign()
        >>> torchhd.cosine_similarity(read, items)
        tensor([[ 1.0000,  0.0469, -0.0117,  0.0039, -0.0313, -0.0078],
                [ 0.0469,  1.0000, -0.0352, -0.0039, -0.0391, -0.0078],
                [-0.0117, -0.0352,  1.0000,  0.0547,  0.0742, -0.0352],
                [ 0.0039, -0.0039,  0.0547,  1.0000,  0.0273,  0.0117],
                [-0.0313, -0.0391,  0.0742,  0.0273,  1.0000, -0.0547],
                [-0.0078, -0.0078, -0.0352,  0.0117, -0.0547,  1.0000]])
    """

    d = query.size(-1)

    query = query.unsqueeze(-2)
    repeat = [1 for _ in range(query.dim())]
    repeat[-2] = d

    pos_query = query.repeat(*repeat)
    torch.diagonal(pos_query, dim1=-2, dim2=-1).fill_(1)

    neg_query = query.repeat(*repeat)
    torch.diagonal(neg_query, dim1=-2, dim2=-1).fill_(-1)

    pos_energy = pos_query @ memory.T
    pos_energy = torch.logsumexp(pos_energy, dim=-1)

    neg_energy = neg_query @ memory.T
    neg_energy = torch.logsumexp(neg_energy, dim=-1)

    return pos_energy - neg_energy


def attention(
    query: Tensor, keys: Tensor, values: Tensor, beta: Optional[float] = None
) -> Tensor:
    r"""`Attention mechanism <https://arxiv.org/abs/1706.03762>`_

    Args:
        query (Tensor): The query vector to compare the similarity with the keys.
        keys (Tensor): The key vectors to compare with the query.
        values (Tensor): The value vectors containing retrievable values from memory.
        beta (float, optional): Temperature scalar for the attention weights before the softmax. Default: 1/sqrt(d)


    Shapes:
        - Query: :math:`(*, f)`
        - Keys: :math:`(n, f)`
        - Values: :math:`(n, g)`
        - Result: :math:`(*, g)`

    Examples::
        >>> items = torchhd.random(6, 512)
        >>> read = torchhd.memory.attention(items, items, items).sign()
        >>> torchhd.cosine_similarity(read, items)
        tensor([[ 1.0000,  0.0625,  0.0117, -0.0625, -0.0078, -0.0430],
                [ 0.0625,  1.0000, -0.0195,  0.0703,  0.0469,  0.0508],
                [ 0.0117, -0.0195,  1.0000,  0.0820,  0.0195,  0.0156],
                [-0.0625,  0.0703,  0.0820,  1.0000, -0.0547, -0.0195],
                [-0.0078,  0.0469,  0.0195, -0.0547,  1.0000, -0.0898],
                [-0.0430,  0.0508,  0.0156, -0.0195, -0.0898,  1.0000]])

    """
    if beta is None:
        d = query.size(-1)
        beta = 1 / math.sqrt(d)

    similarity = query @ keys.T
    scores = torch.softmax(beta * similarity, dim=-1)
    return scores @ values



#
# MIT License
#
# Copyright (c) 2023 Mike Heddes, Igor Nunes, Pere Vergés, Denis Kleyko, and Danny Abraham
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import pytest
import torch
import string

from torchhd import structures, functional
from torchhd.tensors.map import MAPTensor

seed = 2147483645
letters = list(string.ascii_lowercase)


class TestGraph:
    def test_creation_dim(self):
        G = structures.Graph(10000, directed=True)
        assert torch.allclose(G.value, torch.zeros(10000))

    def test_creation_tensor(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random(len(letters), 10000, generator=generator)
        g = functional.bind(hv[0], hv[1])
        G = structures.Graph(g)
        assert torch.allclose(G.value, g)

    def test_generator(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv1 = functional.random(60, 10000, generator=generator)

        generator = torch.Generator()
        generator.manual_seed(seed)
        hv2 = functional.random(60, 10000, generator=generator)

        assert (hv1 == hv2).min().item()

    def test_add_edge(self):
        G = structures.Graph(8)
        hv = torch.tensor(
            [
                [-1.0, 1.0, 1.0, 1.0, -1.0, -1.0, 1.0, -1.0],
                [1.0, -1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [-1.0, -1.0, 1.0, -1.0, -1.0, -1.0, -1.0, 1.0],
                [1.0, 1.0, 1.0, -1.0, 1.0, -1.0, -1.0, -1.0],
            ]
        ).as_subclass(MAPTensor)

        G.add_edge(hv[0], hv[1])
        assert torch.allclose(
            G.value, torch.tensor([-1.0, -1.0, -1.0, 1.0, -1.0, -1.0, 1.0, -1.0])
        )
        G.add_edge(hv[2], hv[3])
        assert torch.allclose(
            G.value, torch.tensor([-2.0, -2.0, 0.0, 2.0, -2.0, 0.0, 2.0, -2.0])
        )

        GD = structures.Graph(8, directed=True)

        GD.add_edge(hv[0], hv[1])
        assert torch.allclose(
            GD.value, torch.tensor([-1.0, 1.0, -1.0, -1.0, -1.0, -1.0, 1.0, -1.0])
        )
        GD.add_edge(hv[2], hv[3])
        assert torch.allclose(
            GD.value, torch.tensor([0.0, 0.0, 0.0, -2.0, 0.0, -2.0, 2.0, -2.0])
        )

    def test_encode_edge(self):
        G = structures.Graph(8)
        hv = torch.tensor(
            [
                [-1.0, 1.0, 1.0, 1.0, -1.0, -1.0, 1.0, -1.0],
                [1.0, -1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [-1.0, -1.0, 1.0, -1.0, -1.0, -1.0, -1.0, 1.0],
                [1.0, 1.0, 1.0, -1.0, 1.0, -1.0, -1.0, -1.0],
            ]
        ).as_subclass(MAPTensor)

        e1 = G.encode_edge(hv[0], hv[1])
        assert torch.allclose(
            e1, torch.tensor([-1.0, -1.0, -1.0, 1.0, -1.0, -1.0, 1.0, -1.0])
        )
        e2 = G.encode_edge(hv[2], hv[3])
        assert torch.allclose(
            e2, torch.tensor([-1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0, -1.0])
        )

        GD = structures.Graph(8, directed=True)

        e1 = GD.encode_edge(hv[0], hv[1])
        assert torch.allclose(
            e1, torch.tensor([-1.0, 1.0, -1.0, -1.0, -1.0, -1.0, 1.0, -1.0])
        )
        e2 = GD.encode_edge(hv[2], hv[3])
        print(e2)
        assert torch.allclose(
            e2, torch.tensor([1.0, -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, -1.0])
        )

    def test_node_neighbors(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random(10, 10000, generator=generator)
        G = structures.Graph(10000, directed=True)

        G.add_edge(hv[0], hv[1])
        G.add_edge(hv[0], hv[2])
        G.add_edge(hv[1], hv[2])

        assert (
            torch.argmax(
                functional.cosine_similarity(G.node_neighbors(hv[1]), hv)
            ).item()
            == 2
        )
        assert functional.cosine_similarity(G.node_neighbors(hv[1]), hv)[2] > 0.5
        assert functional.cosine_similarity(G.node_neighbors(hv[0]), hv)[2] > 0.5
        assert functional.cosine_similarity(G.node_neighbors(hv[0]), hv)[1] > 0.5
        assert functional.cosine_similarity(G.node_neighbors(hv[2]), hv)[1] < 0.5
        assert functional.cosine_similarity(G.node_neighbors(hv[2]), hv)[0] < 0.5
        assert functional.cosine_similarity(G.node_neighbors(hv[1]), hv)[0] < 0.5

        G1 = structures.Graph(10000, directed=False)

        G1.add_edge(hv[0], hv[1])
        G1.add_edge(hv[0], hv[2])
        G1.add_edge(hv[1], hv[2])
        assert functional.cosine_similarity(G1.node_neighbors(hv[1]), hv)[0] > 0.5
        assert functional.cosine_similarity(G1.node_neighbors(hv[0]), hv)[1] > 0.5
        assert functional.cosine_similarity(G1.node_neighbors(hv[0]), hv)[2] > 0.5
        assert functional.cosine_similarity(G1.node_neighbors(hv[2]), hv)[0] > 0.5
        assert functional.cosine_similarity(G1.node_neighbors(hv[1]), hv)[2] > 0.5
        assert functional.cosine_similarity(G1.node_neighbors(hv[2]), hv)[1] > 0.5

    def test_contains(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random(4, 1000, generator=generator)
        G = structures.Graph(1000)

        e1 = G.encode_edge(hv[0], hv[1])
        e2 = G.encode_edge(hv[0], hv[2])
        e3 = G.encode_edge(hv[2], hv[3])

        G.add_edge(hv[0], hv[1])
        G.add_edge(hv[0], hv[2])
        G.add_edge(hv[1], hv[2])

        assert G.contains(e1) > torch.tensor(0.6)
        assert G.contains(e2) > torch.tensor([0.6])
        assert G.contains(e3) < torch.tensor(0.6)

        GD = structures.Graph(1000, directed=True)

        ee1 = GD.encode_edge(hv[0], hv[1])
        ee2 = GD.encode_edge(hv[0], hv[2])
        ee3 = GD.encode_edge(hv[2], hv[3])
        ee4 = GD.encode_edge(hv[1], hv[0])

        GD.add_edge(hv[0], hv[1])
        GD.add_edge(hv[0], hv[2])
        GD.add_edge(hv[3], hv[1])

        assert GD.contains(ee1) > torch.tensor(0.6)
        assert GD.contains(ee2) > torch.tensor(0.6)
        assert GD.contains(ee3) < torch.tensor(0.6)
        assert GD.contains(ee4) < torch.tensor(0.6)

    def test_clear(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random(4, 8, generator=generator)
        G = structures.Graph(8)

        G.add_edge(hv[0], hv[1])
        G.add_edge(hv[0], hv[2])
        G.add_edge(hv[1], hv[2])

        G.clear()

        assert torch.allclose(
            G.value, torch.tensor([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        )

    def test_from_edges(self):
        generator = torch.Generator()
        generator.manual_seed(seed)

        hv = functional.random(4, 1000, generator=generator)
        edges = torch.empty(2, 3, 1000).as_subclass(MAPTensor)
        edges[0, 0] = hv[0]
        edges[1, 0] = hv[1]
        edges[0, 1] = hv[0]
        edges[1, 1] = hv[2]
        edges[0, 2] = hv[1]
        edges[1, 2] = hv[2]

        G = structures.Graph.from_edges(edges)
        neighbors = G.node_neighbors(hv[0])
        neighbor_similarity = functional.cosine_similarity(neighbors, hv)

        assert neighbor_similarity[0] < torch.tensor(0.5)
        assert neighbor_similarity[1] > torch.tensor(0.5)
        assert neighbor_similarity[2] > torch.tensor(0.5)
        assert neighbor_similarity[3] < torch.tensor(0.5)



#
# MIT License
#
# Copyright (c) 2023 Mike Heddes, Igor Nunes, Pere Vergés, Denis Kleyko, and Danny Abraham
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import List, Set, Any
import torch
from torch import Tensor


class VSATensor(Tensor):
    """Base class

    Each model must implement the methods specified on this base class.
    """

    supported_dtypes: Set[torch.dtype]

    @classmethod
    def empty(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=None,
        device=None,
    ) -> "VSATensor":
        """Creates hypervectors representing empty sets"""
        raise NotImplementedError

    @classmethod
    def identity(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=None,
        device=None,
    ) -> "VSATensor":
        """Creates identity hypervectors for binding"""
        raise NotImplementedError

    @classmethod
    def random(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=None,
        device=None,
        generator=None,
    ) -> "VSATensor":
        """Creates random or uncorrelated hypervectors"""
        raise NotImplementedError

    def bundle(self, other: "VSATensor") -> "VSATensor":
        """Bundle the hypervector with other"""
        raise NotImplementedError

    def multibundle(self) -> "VSATensor":
        """Bundle multiple hypervectors"""
        if self.dim() < 2:
            class_name = self.__class__.__name__
            raise RuntimeError(
                f"{class_name} needs to have at least two dimensions for multibundle, got size: {tuple(self.shape)}"
            )

        n = self.size(-2)
        if n == 1:
            return self.unsqueeze(-2)

        tensors: List[VSATensor] = torch.unbind(self, dim=-2)

        output = tensors[0].bundle(tensors[1])
        for i in range(2, n):
            output = output.bundle(tensors[i])

        return output

    def bind(self, other: "VSATensor") -> "VSATensor":
        """Bind the hypervector with other"""
        raise NotImplementedError

    def multibind(self) -> "VSATensor":
        """Bind multiple hypervectors"""
        if self.dim() < 2:
            class_name = self.__class__.__name__
            raise RuntimeError(
                f"{class_name} data needs to have at least two dimensions for multibind, got size: {tuple(self.shape)}"
            )

        n = self.size(-2)
        if n == 1:
            return self.unsqueeze(-2)

        tensors: List[VSATensor] = torch.unbind(self, dim=-2)

        output = tensors[0].bind(tensors[1])
        for i in range(2, n):
            output = output.bind(tensors[i])

        return output

    def inverse(self) -> "VSATensor":
        """Inverse the hypervector for binding"""
        raise NotImplementedError

    def negative(self) -> "VSATensor":
        """Negate the hypervector for the bundling inverse"""
        raise NotImplementedError

    def permute(self, shifts: int = 1) -> "VSATensor":
        """Permute the hypervector"""
        raise NotImplementedError

    def dot_similarity(self, others: "VSATensor") -> Tensor:
        """Inner product with other hypervectors"""
        raise NotImplementedError

    def cosine_similarity(self, others: "VSATensor") -> Tensor:
        """Cosine similarity with other hypervectors"""
        raise NotImplementedError



#
# MIT License
#
# Copyright (c) 2023 Mike Heddes, Igor Nunes, Pere Vergés, Denis Kleyko, and Danny Abraham
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import os.path
from typing import Callable, Optional, Tuple, List
import torch
from torch.utils import data
import pandas as pd
import math

from .utils import download_file_from_google_drive, unzip_file


class EMGHandGestures(data.Dataset):
    """EMG-based hand gestures dataset.

    Dataset from the paper `"Hyperdimensional Biosignal Processing: A Case Study for EMG-based Hand Gesture Recognition" <https://iis-people.ee.ethz.ch/~arahimi/papers/ICRC16.pdf>`_.

    Args:
        root (string): Root directory of dataset where files are stored.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        transform (callable, optional): A function/transform that takes in an torch.FloatTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        subjects (list[int], optional): The subject numbers from 0 til 4 to include. Defaults to [0, 1, 2, 3, 4].
        window (int, optional): The number of measurements to include in each sample. Defaults to 256.
    """

    classes: List[str] = [
        "Closed hand",
        "Open hand",
        "Two-finger pinch",
        "Point index",
        "Rest position",
    ]

    features_files = [
        "COMPLETE_1.csv",
        "COMPLETE_2.csv",
        "COMPLETE_3.csv",
        "COMPLETE_4.csv",
        "COMPLETE_5.csv",
    ]

    labels_files = [
        "LABEL_1.csv",
        "LABEL_2.csv",
        "LABEL_3.csv",
        "LABEL_4.csv",
        "LABEL_5.csv",
    ]

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
        subjects: list = [0, 1, 2, 3, 4],
        window: int = 256,
    ):
        root = os.path.join(root, "EMG_based_hand_gesture")
        root = os.path.expanduser(root)
        self.root = root
        os.makedirs(self.root, exist_ok=True)

        self.transform = transform
        self.target_transform = target_transform
        self.subjects = subjects
        self.window = window
        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it"
            )

        self._load_data()

    def __len__(self) -> int:
        return self.data.size(0)

    def __getitem__(self, index: int) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """
        Args:
            index (int): Index

        Returns:
            Tuple[torch.FloatTensor, torch.LongTensor]: (sample, target) where target is the index of the target class
        """
        sample = self.data[index]
        label = self.targets[index]

        if self.transform:
            sample = self.transform(sample)

        if self.target_transform:
            label = self.target_transform(label)

        return sample, label

    def _check_integrity(self) -> bool:
        if not os.path.isdir(self.root):
            return False

        has_not_feature_files = sum(
            list(
                map(
                    lambda x: not os.path.isfile(os.path.join(self.root, x)),
                    self.features_files,
                )
            )
        )
        has_not_label_files = sum(
            list(
                map(
                    lambda x: not os.path.isfile(os.path.join(self.root, x)),
                    self.labels_files,
                )
            )
        )
        # Check if the root directory contains the required files
        if not has_not_feature_files and not has_not_label_files:
            return True
        return False

    def _load_data(self):
        features = torch.empty(0, dtype=torch.long)
        labels = torch.empty(0, dtype=torch.long)
        for i in self.subjects:
            complete = pd.read_csv(
                os.path.join(self.root, self.features_files[i]), header=None
            )
            label = pd.read_csv(
                os.path.join(self.root, self.labels_files[i]), header=None
            )
            # List of indices where the gesture changes
            indexes = [
                index
                for index, _ in enumerate(label.values)
                if label.values[index] != label.values[index - 1]
            ]
            prev = 0
            labels_clean = torch.empty(0, dtype=torch.long)
            features_clean = torch.empty(0, self.window, 4, dtype=torch.long)
            # Every change of gesture we group it values
            for j in indexes:
                span = j - prev
                # If we have that the amount of data of the gesture fits in the window we have a new sample of the gesture of size window
                if span > self.window:
                    for k in range(math.floor(span / self.window)):
                        # Clean the label data
                        label_clean = (
                            torch.tensor(
                                label.values[prev + (self.window * k)], dtype=torch.long
                            )
                            - 1
                        )
                        # Clean the feature data
                        feature_clean = torch.tensor(
                            complete.values[
                                prev
                                + (self.window * k) : prev
                                + (self.window * (k + 1))
                            ],
                            dtype=torch.long,
                        )[None, :, :]
                        labels_clean = torch.cat((labels_clean, label_clean))
                        features_clean = torch.cat((features_clean, feature_clean))
                prev = j
            features = torch.cat((features, features_clean))
            labels = torch.cat((labels, labels_clean))
        self.data = features
        self.targets = labels

    def download(self):
        """Download the data if it doesn't exist already."""

        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        zip_file_path = os.path.join(self.root, "data.zip")
        download_file_from_google_drive(
            "1_9DYP1MwICtnMRm_F34xVvv1S2OhVnHH",  # Google Drive shared file ID
            zip_file_path,
        )

        unzip_file(zip_file_path, self.root)
        os.remove(zip_file_path)



#
# MIT License
#
# Copyright (c) 2023 Mike Heddes, Igor Nunes, Pere Vergés, Denis Kleyko, and Danny Abraham
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import os.path
from typing import Callable, Optional, NamedTuple, List
import torch
from torch.utils import data
import pandas as pd

from .utils import download_file, unzip_file


class BeijingAirQualityDataSample(NamedTuple):
    categorical: torch.LongTensor
    continuous: torch.FloatTensor


class BeijingAirQuality(data.Dataset):
    """`Beijing Multi-Site Air-Quality <https://archive.ics.uci.edu/ml/datasets/Beijing+Multi-Site+Air-Quality+Data>`_ dataset.

    .. list-table::
       :widths: 10 10 10 10
       :align: center
       :header-rows: 1

       * - Instances
         - Attributes
         - Task
         - Area
       * - 420768
         - 18
         - Regression
         - Physical

    .. warning::
        The data contains NaN values that need to be taken into account.

    Args:
        root (string): Root directory of dataset where directory
            ``beijing-air-quality`` exists or will be saved to if download is set to True.
        transform (callable, optional): A function/transform that takes in a feature Tensor
            and returns a transformed version.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.

    """

    wind_directions: List[str]
    stations: List[str]
    categorical_columns: List[str] = [
        "year",
        "month",
        "day",
        "hour",
        "wind_direction",
        "station",
    ]
    continuous_columns: List[str] = [
        "PM2.5",
        "PM10",
        "SO2",
        "NO2",
        "CO",
        "O3",
        "temperature",
        "pressure",
        "dew_point_temperature",
        "precipitation",
        "wind_speed",
    ]

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        download: bool = False,
    ):
        root = os.path.join(root, "beijing-air-quality")
        root = os.path.expanduser(root)
        self.root = root
        os.makedirs(self.root, exist_ok=True)

        self.transform = transform

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it"
            )

        self._load_data()

    def __len__(self) -> int:
        return self.categorical_data.size(0)

    def __getitem__(self, index: int) -> BeijingAirQualityDataSample:
        """
        Args:
            index (int): Index

        Returns:
            BeijingAirQualityDataSample: Indexed Sample
        """
        sample = BeijingAirQualityDataSample(
            self.categorical_data[index], self.continuous_data[index]
        )

        if self.transform:
            sample = self.transform(sample)

        return sample

    def _check_integrity(self) -> bool:
        if not os.path.isdir(self.root):
            return False

        # Check if the root directory contains the expected number of csv files
        files = os.listdir(self.root)
        files = [file for file in files if file.endswith(".csv")]
        if len(files) == 12:
            return True

        # TODO: Add more specific checks like an MD5 checksum

        return False

    def _load_data(self):
        files = os.listdir(self.root)
        csv_files = [file for file in files if file.endswith(".csv")]

        data_tables = []
        for filename in csv_files:
            data = pd.read_csv(os.path.join(self.root, filename))
            # No column does not provide any meaningful information
            data = data.drop(columns=["No"])
            data_tables.append(data)

        data = pd.concat(data_tables, ignore_index=True)
        # Rename columns for accessability
        data = data.rename(
            columns={
                "wd": "wind_direction",
                "TEMP": "temperature",
                "PRES": "pressure",
                "DEWP": "dew_point_temperature",
                "RAIN": "precipitation",
                "WSPM": "wind_speed",
            }
        )

        # Change null values to Not Available string "N/A"
        data.loc[data.wind_direction.isnull(), "wind_direction"] = "N/A"
        # Map the wind directions to a category identifier (int)
        # # Save the mapping for user referencing
        self.wind_directions = tuple(sorted(list(set(data.wind_direction))))
        data.wind_direction = data.wind_direction.apply(
            lambda x: self.wind_directions.index(x)
        )

        # Map the stations to a category identifier (int)
        # Save the mapping for user referencing
        self.stations = tuple(sorted(list(set(data.station))))
        data.station = data.station.apply(lambda x: self.stations.index(x))

        categorical = data[self.categorical_columns]
        self.categorical_data = torch.tensor(categorical.values, dtype=torch.long)

        continuous = data[self.continuous_columns]
        self.continuous_data = torch.tensor(continuous.values, dtype=torch.float)

    def download(self):
        """Download the data if it doesn't exist already."""

        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        zip_file_path = os.path.join(self.root, "data.zip")
        download_file(
            "https://archive.ics.uci.edu/ml/machine-learning-databases/00501/PRSA2017_Data_20130301-20170228.zip",
            zip_file_path,
        )

        unzip_file(zip_file_path, self.root)
        os.remove(zip_file_path)

        source_dir = os.path.join(self.root, "PRSA_Data_20130301-20170228")
        data_files = os.listdir(source_dir)
        for filename in data_files:
            os.rename(
                os.path.join(source_dir, filename), os.path.join(self.root, filename)
            )

        os.rmdir(source_dir)
